import sys
from pathlib import Path

from aicage.constants import DEFAULT_IMAGE_BASE

_OS_RELEASE_PATH: Path = Path("/etc/os-release")


def resolve_default_base(available_bases: list[str]) -> str:
    if not available_bases:
        return DEFAULT_IMAGE_BASE
    for candidate in _preferred_bases_for_host():
        matched = _find_matching_base(candidate, available_bases)
        if matched is not None:
            return matched
    if DEFAULT_IMAGE_BASE in available_bases:
        return DEFAULT_IMAGE_BASE
    return available_bases[0]


def _preferred_bases_for_host() -> list[str]:
    if not sys.platform.startswith("linux"):
        return [DEFAULT_IMAGE_BASE]
    return _host_distro_tokens()


def _host_distro_tokens() -> list[str]:
    data = _read_os_release()
    tokens: list[str] = []
    distro_id = data.get("id")
    id_like = data.get("id_like")
    if distro_id:
        token = distro_id.strip().lower()
        if token:
            tokens.append(token)
    if id_like:
        for raw_token in id_like.split():
            token = raw_token.strip().lower()
            if token:
                tokens.append(token)
    return tokens


def _read_os_release() -> dict[str, str]:
    try:
        content = _OS_RELEASE_PATH.read_text(encoding="utf-8")
    except OSError:
        return {}
    values: dict[str, str] = {}
    for raw_line in content.splitlines():
        line = raw_line.strip()
        if not line or line.startswith("#") or "=" not in line:
            continue
        key, _, value = line.partition("=")
        key = key.strip().lower()
        value = value.strip().strip("\"'")
        if key:
            values[key] = value.lower()
    return values


def _find_matching_base(candidate: str, available_bases: list[str]) -> str | None:
    candidate_lower = candidate.lower()
    for base in available_bases:
        if base.lower() == candidate_lower:
            return base
    return None



import json
from collections.abc import Callable
from functools import lru_cache
from typing import Any

from jsonschema import Draft202012Validator
from jsonschema.exceptions import ValidationError

from aicage.config.errors import ConfigError
from aicage.config.resources import find_packaged_path

_Normalizer = Callable[[dict[str, Any]], dict[str, Any]]


@lru_cache(maxsize=1)
def load_schema(path: str) -> dict[str, Any]:
    payload = find_packaged_path(path).read_text(encoding="utf-8")
    return json.loads(payload)


def validate_schema_mapping(
    mapping: dict[str, Any],
    schema: dict[str, Any],
    context: str,
    *,
    normalizer: _Normalizer | None = None,
) -> dict[str, Any]:
    if not isinstance(mapping, dict):
        raise ConfigError(f"{context} must be a mapping.")

    normalized = dict(mapping)
    if normalizer is not None:
        normalized = normalizer(normalized)

    validator = Draft202012Validator(schema)
    errors = sorted(validator.iter_errors(normalized), key=lambda error: list(error.path))
    if errors:
        raise ConfigError(_error_message(context, errors[0]))
    return normalized


def _error_message(context: str, error: ValidationError) -> str:
    if error.path:
        path = ".".join(str(token) for token in error.path)
        return f"{context}.{path} {error.message}"
    return f"{context} {error.message}"



from collections.abc import Callable
from typing import TypeVar

_T = TypeVar("_T")


def maybe_build(
    load_record: Callable[[], _T | None],
    should_rebuild: Callable[[_T | None], bool],
    run_build: Callable[[], None],
    save_record: Callable[[], object],
) -> bool:
    record = load_record()
    if not should_rebuild(record):
        return False
    run_build()
    save_record()
    return True



from aicage.cli_types import ParsedArgs
from aicage.config.context import ConfigContext
from aicage.config.project_config import AgentConfig
from aicage.runtime.docker_args._support._resolver_types import MountRequest, ResolvedArgs

from .._support._git_support import resolve_git_config_path


def resolve(
    context: ConfigContext,
    agent: str,
    parsed: ParsedArgs | None,
) -> ResolvedArgs:
    _ = parsed
    agent_cfg: AgentConfig = context.project_cfg.agents[agent]
    git_config = resolve_git_config_path()
    if not git_config or not git_config.exists():
        return ResolvedArgs()

    mounts_cfg = agent_cfg.mounts
    if mounts_cfg.gitconfig:
        return ResolvedArgs(mounts=[MountRequest(host_path=git_config)])
    return ResolvedArgs()



import os
from pathlib import Path

from aicage.paths import container_project_path
from aicage.runtime.env_vars import AICAGE_GID, AICAGE_HOME, AICAGE_HOST_IS_LINUX, AICAGE_HOST_USER, AICAGE_UID


def resolve_user_ids() -> list[str]:
    env_flags: list[str] = []
    host_home = _resolve_host_home()
    host_home_posix = container_project_path(host_home).as_posix()
    if os.name == "nt":
        env_flags.extend(["-e", f"{AICAGE_HOME}={host_home_posix}"])
        return env_flags

    getuid = getattr(os, "getuid", None)
    getgid = getattr(os, "getgid", None)
    uid = getuid() if callable(getuid) else None
    gid = getgid() if callable(getgid) else None

    user = _resolve_host_user()
    if uid is not None:
        env_flags.extend(["-e", f"{AICAGE_UID}={uid}", "-e", f"{AICAGE_GID}{'='}{gid}"])
    env_flags.extend(["-e", f"{AICAGE_HOST_USER}={user}"])
    env_flags.extend(
        [
            "-e",
            f"{AICAGE_HOME}={host_home_posix}",
            "-e",
            f"{AICAGE_HOST_IS_LINUX}=true",
        ]
    )
    return env_flags


def _resolve_host_user() -> str:
    return os.environ.get("USERNAME") or os.environ.get("USER") or "aicage"


def _resolve_host_home() -> Path:
    return Path.home()



from dataclasses import dataclass
from pathlib import Path, PureWindowsPath

from aicage.errors import AicageError


@dataclass(frozen=True)
class ShareSpec:
    host_path: Path
    read_only: bool


def resolve_share_specs(
    raw_shares: list[str],
    cwd: Path,
) -> list[ShareSpec]:
    share_mounts: list[ShareSpec] = []
    if not raw_shares:
        return share_mounts
    mounted_hosts: set[Path] = set()
    for raw in raw_shares:
        share = _parse_share(raw, cwd)
        if share.host_path in mounted_hosts:
            continue
        mounted_hosts.add(share.host_path)
        share_mounts.append(ShareSpec(host_path=share.host_path, read_only=share.read_only))
    return share_mounts


def merge_share_values(
    cli_shares: list[str],
    existing_shares: list[str],
    cwd: Path,
) -> tuple[list[str], list[str]]:
    cli_specs = _normalize_share_specs(cli_shares, cwd)
    existing_specs = _normalize_share_specs(existing_shares, cwd)
    existing_hosts = {spec.host_path for spec in existing_specs}
    cli_hosts: set[Path] = set()
    merged: list[str] = []
    new_shares: list[str] = []
    for spec in cli_specs:
        if spec.host_path in cli_hosts:
            continue
        share_value = _format_share_value(spec)
        cli_hosts.add(spec.host_path)
        merged.append(share_value)
        if spec.host_path not in existing_hosts:
            new_shares.append(share_value)
    for spec in existing_specs:
        if spec.host_path in cli_hosts:
            continue
        merged.append(_format_share_value(spec))
    return merged, new_shares


def _parse_share(raw: str, cwd: Path) -> ShareSpec:
    if not raw:
        raise AicageError("Share value cannot be empty.")
    read_only, host_raw = _parse_read_only(raw)
    if _has_destination_path(host_raw):
        raise AicageError("Share destinations are not supported. Use HOST or HOST:ro.")
    host_path = _resolve_host_path(host_raw, cwd)
    if host_path.exists() and not host_path.is_file() and not host_path.is_dir():
        raise AicageError(f"Path exists but is not a file or directory: {host_path}")
    return ShareSpec(host_path=host_path, read_only=read_only)


def _parse_read_only(raw: str) -> tuple[bool, str]:
    if raw.endswith(":ro"):
        host_raw = raw[:-3]
        if not host_raw:
            raise AicageError("Share host path cannot be empty.")
        return True, host_raw
    return False, raw


def _resolve_host_path(host_raw: str, cwd: Path) -> Path:
    host_path = Path(host_raw).expanduser()
    if not host_path.is_absolute():
        host_path = cwd / host_path
    return host_path.resolve()


def _has_destination_path(host_raw: str) -> bool:
    if ":" not in host_raw:
        return False
    windows_path = PureWindowsPath(host_raw)
    if windows_path.drive and host_raw.startswith(windows_path.drive):
        return host_raw.count(":") > 1
    return True


def _normalize_share_specs(raw_shares: list[str], cwd: Path) -> list[ShareSpec]:
    specs: list[ShareSpec] = []
    seen: set[Path] = set()
    for raw in raw_shares:
        spec = _parse_share(raw, cwd)
        if spec.host_path in seen:
            continue
        seen.add(spec.host_path)
        specs.append(spec)
    return specs


def _format_share_value(spec: ShareSpec) -> str:
    host_value = str(spec.host_path)
    return f"{host_value}:ro" if spec.read_only else host_value



"""Shared environment variable names used for container setup."""

AICAGE_UID = "AICAGE_UID"
AICAGE_GID = "AICAGE_GID"
AICAGE_HOST_USER = "AICAGE_HOST_USER"
AICAGE_HOME = "AICAGE_HOME"
AICAGE_HOST_IS_LINUX = "AICAGE_HOST_IS_LINUX"
AICAGE_WORKSPACE = "AICAGE_WORKSPACE"

DOCKER_HOST = "DOCKER_HOST"
WINDOWS_DOCKER_HOST = "tcp://host.docker.internal:2375"



from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path

from aicage.config.extensions.loader import ExtensionMetadata
from aicage.config.project_config import AgentConfig
from aicage.runtime.mounts.shares import ShareSpec, resolve_share_specs
from aicage.runtime.prompts.confirm import prompt_mount_git_support

from ._git_support import git_support_prompt_items


@dataclass(frozen=True)
class _ExtensionMountPromptItem:
    extension_id: str
    description: str


@dataclass(frozen=True)
class _MountPromptPrefs:
    git_mounts: set[str]
    extension_mounts: dict[str, bool]


def resolve_mount_prompt_prefs(
    project_path: Path,
    agent_cfg: AgentConfig,
    available_extensions: dict[str, ExtensionMetadata],
) -> _MountPromptPrefs | None:
    git_items = git_support_prompt_items(project_path, agent_cfg.mounts)
    extension_items = _extension_mount_prompt_items(
        project_path,
        agent_cfg.extensions,
        agent_cfg.extension_mounts,
        available_extensions,
    )
    if not git_items and not extension_items:
        return None

    selected = set(
        prompt_mount_git_support(
            list(git_items),
            _format_extension_prompt_items(extension_items),
        )
    )
    return _MountPromptPrefs(
        git_mounts={key for key, _description in git_items if key in selected},
        extension_mounts={
            item.extension_id: item.extension_id in selected
            for item in extension_items
        },
    )


def _format_extension_prompt_items(
    extension_items: Iterable[_ExtensionMountPromptItem],
) -> list[tuple[str, str]]:
    return [(item.extension_id, item.description) for item in extension_items]


def _extension_mount_prompt_items(
    project_path: Path,
    selected_extensions: list[str],
    extension_mounts: dict[str, bool],
    available_extensions: dict[str, ExtensionMetadata],
) -> list[_ExtensionMountPromptItem]:
    items: list[_ExtensionMountPromptItem] = []
    for extension_id in selected_extensions:
        if extension_id in extension_mounts:
            continue
        extension = available_extensions.get(extension_id)
        if extension is None or not extension.shares:
            continue
        share_specs = resolve_share_specs(extension.shares, project_path)
        share_values = ", ".join(_format_share_value(spec) for spec in share_specs)
        items.append(
            _ExtensionMountPromptItem(
                extension_id=extension_id,
                description=f"Extension {extension_id} shares: {share_values}",
            )
        )
    return items


def _format_share_value(spec: ShareSpec) -> str:
    host_value = str(spec.host_path)
    return f"{host_value}:ro" if spec.read_only else host_value



from pathlib import Path

from aicage.cli_types import ParsedArgs
from aicage.config.context import ConfigContext
from aicage.config.project_config import AgentConfig
from aicage.runtime.docker_args._support._resolver_types import MountRequest, ResolvedArgs

from .._support._git_support import resolve_ssh_dir, uses_ssh_remotes
from .._support._signing import is_commit_signing_enabled, resolve_signing_format


def resolve(
    context: ConfigContext,
    agent: str,
    parsed: ParsedArgs | None,
) -> ResolvedArgs:
    _ = parsed
    project_path = Path(context.project_cfg.path)
    agent_cfg: AgentConfig = context.project_cfg.agents[agent]
    signing_enabled = is_commit_signing_enabled(project_path)
    signing_format = resolve_signing_format(project_path) if signing_enabled else None
    ssh_needed = (signing_enabled and signing_format == "ssh") or uses_ssh_remotes(project_path)
    if not ssh_needed:
        return ResolvedArgs()

    ssh_dir = resolve_ssh_dir()
    if not ssh_dir.exists():
        return ResolvedArgs()

    mounts_cfg = agent_cfg.mounts
    if mounts_cfg.ssh:
        return ResolvedArgs(mounts=[MountRequest(host_path=ssh_dir)])
    return ResolvedArgs()



def repository_from_image_ref(image_ref: str) -> str:
    name = image_ref.split("@", 1)[0]
    last_colon = name.rfind(":")
    if last_colon > name.rfind("/"):
        return name[:last_colon]
    return name



from pathlib import Path

from aicage._logging import get_logger
from aicage.config.config_store import SettingsStore


def remove_project_config(agent: str | None = None) -> None:
    logger = get_logger()
    store = SettingsStore()
    project_path = Path.cwd().resolve()
    config_path = store.project_config_path(project_path)
    if agent is not None:
        _remove_agent_config(store, project_path, config_path, agent)
        return
    logger.info("Removing project config at %s", config_path)
    if config_path.exists():
        config_path.unlink()
        print("Project config removed:")
        print(config_path)
    else:
        print("Project config not found:")
        print(config_path)


def _remove_agent_config(
    store: SettingsStore,
    project_path: Path,
    config_path: Path,
    agent: str,
) -> None:
    logger = get_logger()
    logger.info("Removing agent config '%s' from %s", agent, config_path)
    if not config_path.exists():
        print("Project config not found:")
        print(config_path)
        return
    config = store.load_project(project_path)
    if agent not in config.agents:
        print("Agent config not found:")
        print(agent)
        return
    del config.agents[agent]
    store.save_project(project_path, config)
    print("Agent config removed:")
    print(agent)
    print("Project config updated:")
    print(config_path)



from typing import Any

from aicage.config._schema_validation import load_schema, validate_schema_mapping
from aicage.config.base.models import BUILD_LOCAL_KEY

_BASE_SCHEMA_PATH: str = "validation/base.schema.json"
_BASE_CONTEXT: str = "base metadata"


def validate_base_mapping(mapping: dict[str, Any]) -> dict[str, Any]:
    schema = load_schema(_BASE_SCHEMA_PATH)
    return validate_schema_mapping(
        mapping,
        schema,
        _BASE_CONTEXT,
        normalizer=_apply_defaults,
    )


def _apply_defaults(mapping: dict[str, Any]) -> dict[str, Any]:
    normalized = dict(mapping)
    normalized.setdefault(BUILD_LOCAL_KEY, False)
    return normalized



from aicage._logging import get_logger
from aicage.docker.pull import run_pull
from aicage.docker.query import cleanup_old_digest, get_local_repo_digest_for_repo
from aicage.registry._errors import RegistryError
from aicage.registry._logs import pull_log_path
from aicage.registry._signature import resolve_verified_digest


def refresh_base_digest(
    base_image_ref: str,
    base_repository: str,
) -> str:
    logger = get_logger()
    local_digest = get_local_repo_digest_for_repo(base_image_ref, base_repository)
    try:
        digest_ref = resolve_verified_digest(base_image_ref)
    except RegistryError:
        if not local_digest:
            raise
        logger.warning("Base image digest check failed; using local base image.")
        return f"{base_repository}@{local_digest}"
    remote_digest = digest_ref.split("@", 1)[1]
    if remote_digest == local_digest:
        return digest_ref

    log_path = pull_log_path(base_image_ref)
    try:
        run_pull(base_image_ref, log_path)
    except RegistryError:
        if local_digest:
            logger.warning(
                "Base image pull failed; using local base image (logs: %s).", log_path
            )
            return f"{base_repository}@{local_digest}"
        raise

    cleanup_old_digest(base_repository, local_digest, base_image_ref)
    return digest_ref



import argparse
import sys
from collections.abc import Sequence

from aicage import __version__
from aicage._logging import get_logger
from aicage.cli._errors import CliError
from aicage.cli_types import ParsedArgs

_MIN_REMAINING_WITH_AGENT = 2
_MAX_CONFIG_TOKENS = 2
_CONFIG_ACTION_ALIASES: dict[str, str] = {
    "print": "info",
}
_VALID_CONFIG_ACTIONS: set[str] = {"info", "remove"}


def parse_cli(argv: Sequence[str]) -> ParsedArgs:
    """
    Returns parsed CLI args.
    Docker args are captured as an opaque string; precedence is resolved later.
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--dry-run", action="store_true", help="Print docker run command without executing.")
    parser.add_argument("-y", "--yes", action="store_true", help="Use defaults for all prompts.")
    parser.add_argument("--docker", action="store_true", help="Mount the host Docker socket into the container.")
    parser.add_argument(
        "--share",
        action="append",
        default=[],
        help="Mount a host directory into the container (repeatable).",
    )
    parser.add_argument(
        "--config",
        nargs="*",
        help="Perform config actions such as the default info or 'remove [agent]'.",
    )
    parser.add_argument("-h", "--help", action="store_true", help="Show help message and exit.")
    pre_argv, post_argv = _split_argv(argv)

    opts: argparse.Namespace
    remaining: list[str]
    if len(pre_argv) == 1 and pre_argv[0] in ("-v", "--version") and post_argv is None:
        print(__version__)
        get_logger().info("Displayed CLI version.")
        sys.exit(0)

    opts, remaining = parser.parse_known_args(pre_argv)

    if opts.help:
        usage: str = (
            "Usage:\n"
            "  aicage <agent>\n"
            "  aicage [--dry-run] [--docker] [--share <path>...] <agent> [<agent-args>]\n"
            "  aicage [--dry-run] [--docker] [--share <path>...] <docker-args> -- <agent> [<agent-args>]\n"
            "  aicage -y [--dry-run] [--docker] [--share <path>...] <agent> [<agent-args>]\n"
            "  aicage --config\n"
            "  aicage --config info\n"
            "  aicage --config remove [<agent>]\n"
            "  aicage --version\n\n"
            "Arguments:\n"
            "  --dry-run        Print the generated docker run command and exit.\n"
            "  -y, --yes        Use default answers for all prompts.\n"
            "  --docker         Mount /var/run/docker.sock into the container.\n"
            "  --share <path>   Mount a host path into the container. Repeatable.\n"
            "  --config [<cmd>] Run config command: default info, or remove [agent].\n"
            "  -v, --version    Print aicage version and exit.\n"
            "  -h, --help       Show this help and exit.\n\n"
            "Behavior:\n"
            "  - <docker-args> are forwarded verbatim to docker run.\n"
            "  - If docker args are present, use '--' before <agent>.\n"
            "  - <agent-args> are forwarded verbatim to the agent.\n"
        )
        print(usage)
        get_logger().info("Displayed CLI usage help.")
        sys.exit(0)

    if opts.config is not None:
        config_tokens: list[str] = opts.config
        config_action, config_agent = _validate_config_action(config_tokens, opts, remaining, post_argv)
        return ParsedArgs(
            opts.dry_run,
            "",
            "",
            [],
            opts.docker,
            opts.share,
            config_action,
            config_agent,
            opts.yes,
        )

    docker_args, agent, agent_args = _parse_agent_section(remaining, post_argv)

    if not agent:
        raise CliError("Agent name is required.")

    return ParsedArgs(
        opts.dry_run,
        docker_args,
        agent,
        agent_args,
        opts.docker,
        opts.share,
        None,
        None,
        opts.yes,
    )


def _split_argv(argv: Sequence[str]) -> tuple[list[str], list[str] | None]:
    if "--" not in argv:
        return list(argv), None
    sep_index = argv.index("--")
    pre_argv = list(argv[:sep_index])
    post_argv = list(argv[sep_index + 1 :])
    return pre_argv, post_argv


def _normalize_config_action(action: str) -> str:
    return _CONFIG_ACTION_ALIASES.get(action, action)


def _validate_config_action(
    config_tokens: list[str],
    opts: argparse.Namespace,
    remaining: list[str],
    post_argv: list[str] | None,
) -> tuple[str, str | None]:
    if not config_tokens:
        config_tokens = ["info"]
    if len(config_tokens) > _MAX_CONFIG_TOKENS:
        raise CliError("Too many values for --config. Use '--config remove [agent]'.")
    raw_action = config_tokens[0]
    config_action = _normalize_config_action(raw_action)
    if config_action not in _VALID_CONFIG_ACTIONS:
        raise CliError(f"Unknown config action: {raw_action}")
    config_agent = config_tokens[1] if len(config_tokens) == _MAX_CONFIG_TOKENS else None
    if config_action == "info" and config_agent is not None:
        raise CliError("No agent value is allowed with '--config info'.")
    if remaining or post_argv or opts.docker or opts.dry_run or opts.share or opts.yes:
        raise CliError("No additional arguments are allowed with --config.")
    return config_action, config_agent


def _parse_agent_section(
    remaining: list[str],
    post_argv: list[str] | None,
) -> tuple[str, str, list[str]]:
    if post_argv is not None:
        if not post_argv:
            raise CliError("Missing agent after '--'.")
        docker_args = " ".join(remaining).strip()
        return docker_args, post_argv[0], post_argv[1:]
    if not remaining:
        raise CliError("Missing arguments. Provide an agent name (and optional docker args).")
    first: str = remaining[0]
    if first.startswith("-") or "=" in first:
        raise CliError("Docker args require '--' before the agent.")
    return "", first, remaining[1:]



from pathlib import Path

import yaml

from aicage import paths as paths_module
from aicage.registry._time import now_iso

_AGENT_KEY: str = "agent"
_VERSION_KEY: str = "version"
_CHECKED_AT_KEY: str = "checked_at"


class VersionCheckStore:
    def __init__(self) -> None:
        self._base_dir = paths_module.AGENT_VERSION_CHECK_STATE_DIR

    def load(self, agent: str) -> str | None:
        path = self._base_dir / f"{_sanitize_agent_name(agent)}.yml"
        if not path.is_file():
            return None
        with path.open("r", encoding="utf-8") as handle:
            payload = yaml.safe_load(handle)
        if not isinstance(payload, dict):
            return None
        version = payload.get(_VERSION_KEY)
        if not isinstance(version, str) or not version:
            return None
        return version

    def save(self, agent: str, version: str) -> Path:
        self._base_dir.mkdir(parents=True, exist_ok=True)
        path = self._base_dir / f"{_sanitize_agent_name(agent)}.yml"
        with path.open("w", encoding="utf-8") as handle:
            payload = {
                _AGENT_KEY: agent,
                _VERSION_KEY: version,
                _CHECKED_AT_KEY: now_iso(),
            }
            yaml.safe_dump(payload, handle, sort_keys=True)
        return path


def _sanitize_agent_name(agent_name: str) -> str:
    return agent_name.replace("/", "_")
